' ---------------------------------------------
' base collision detection routines
' ---------------------------------------------

' abstract for base collision detection
Type ICOLLISION_DETECT Abstract
	Field _method:Int
	Field _response:ICOLLISION_RESPONSE=Null
	
	' returns NULL if no collision
	Method GetCollision:COLLISION(entity1:ib3d_Entity,entity2:ib3d_Entity)
		Return Null
	EndMethod
	
	Method setCollisionResponse(response:ICOLLISION_RESPONSE)
		_response=response
	EndMethod
EndType

' Irrlicht-based collision detection with response criteria
Type ICOLLISION_DETECLICHT Extends ICOLLISION_DETECT Abstract
	Field _slide:Float=0.00000000
	Field _gravity:Vector3df=Vector3df.Create()
	
	Method setGravity(gravity:Vector3df)
		If gravity And gravity.IsValid() Then _gravity=gravity Else _gravity=Vector3df.createFromVals(0,0,0)
	EndMethod
	
	' returns NULL if no collision
	Method GetCollision:COLLISION(entity1:ib3d_Entity,entity2:ib3d_Entity)
		Return Null
	EndMethod
	
	Method setCollisionResponse(response:ICOLLISION_RESPONSE)
		Super.setCollisionResponse(response)
		If ICOLLISION_RESPONSE_IRRLICHT(response)
			_gravity=ICOLLISION_RESPONSE_IRRLICHT(response)._gravity
			_slide=ICOLLISION_RESPONSE_IRRLICHT(response)._slide			
		EndIf
	EndMethod
EndType

' detects if 2 spheres intersect using irrlicht as the source of detection
Type DETECT_SPHERE_SPHERE_IRRLICHT Extends ICOLLISION_DETECLICHT
	
	Function Create:DETECT_SPHERE_SPHERE_IRRLICHT()
		Local retval:DETECT_SPHERE_SPHERE_IRRLICHT=New DETECT_SPHERE_SPHERE_IRRLICHT
		retval._method=CD_SPHERE_SPHERE_IRRLICHT
		Return retval
	EndFunction
	
	' this is a sphere-shpere collision
	Method GetCollision:COLLISION(entity1:ib3d_Entity,entity2:ib3d_Entity)
		Local retval:COLLISION=Null

		Local pos:Vector3df=entity1._node.getPosition()
		
		' if there is an offset, add it in to the position
		If entity1._collision_offset Then pos.PlusEq(entity1._collision_offset)
				
		' if the distance between is less than the sum of the radius', we have a collision
		If pos.getDistanceFrom(entity2._node.getPosition())<(entity1._radius.getX()+entity2._radius.getX())
			Return COLLISION.Create(entity1,entity2,_method)
		EndIf
		
		Return retval
	EndMethod
EndType

' detects if an ellipsoid intersects with a triangle of a static polygon
Type DETECT_ELLIPSOID_POLY_STATIC_IRRLICHT Extends ICOLLISION_DETECLICHT
	
	Function Create:DETECT_ELLIPSOID_POLY_STATIC_IRRLICHT()
		Local retval:DETECT_ELLIPSOID_POLY_STATIC_IRRLICHT=New DETECT_ELLIPSOID_POLY_STATIC_IRRLICHT
		retval._method=CD_ELLIPSOID_POLY_STATIC_IRRLICHT
		Return retval
	EndFunction
	
	' this is a ellipsoid-poly collision
	Method GetCollision:COLLISION(entity1:ib3d_Entity,entity2:ib3d_Entity)
		If Not MESH(entity2) Then RuntimeError "Cannot detect ELLIPSOID to POLY STATIC collisions unless checking against a MESH"
		If Not MESH(entity2)._mesh Then RuntimeError "Invalid MESH"
		
		Local retval:COLLISION=Null
		Local vel:Vector3df=entity1._node.getPosition().Minus(entity1.getLastPosition())

		' make sure we are moving
		If vel.Plus(_gravity).NotEq(Vector3df.createFromVals(0,0,0))
			' for ellipsoid-poly static, create the triangle selector once.  this should be static geometry that doesnt move.
			Local triselector:ITriangleSelector=entity2._node.getTriangleSelector()
			If Not triselector Or Not triselector.IsValid()
				If IAnimatedMesh(MESH(entity2)._mesh)
					triselector=_g_ib3d_engine.smgr.createOctreeTriangleSelector(IAnimatedMesh(MESH(entity2)._mesh).getMesh(0),entity2._node)
				Else
					triselector=_g_ib3d_engine.smgr.createOctreeTriangleSelector(IMesh(MESH(entity2)._mesh),entity2._node)
				EndIf
				entity2._node.setTriangleSelector(triselector)
			EndIf 
			
			Local pos:Vector3df		
			Local tri:Triangle3df=Triangle3df.Create()
			Local outPoint:Vector3df=Vector3df.Create()
			Local falling:Int=False
			Local outNode:ISceneNode
					
			pos=_g_ib3d_engine.smgr.getSceneCollisionManager().getCollisionResultPosition(triselector,entity1.getLastPosition().Minus(entity1._collision_offset),entity1._radius,vel,tri,outPoint,Varptr(falling),outNode,_slide,_gravity)
										
			pos.PlusEq(entity1._collision_offset)
			outPoint.PlusEq(entity1._node.getPosition())
			outPoint.PlusEq(entity1._collision_offset)		

			If falling Or (tri.getPointA().getX()<>0 Or tri.getPointA().getY()<>0 Or tri.getPointA().getZ()<>0)
				Return COLLISION.Create(entity1,entity2,_method,tri,pos,falling,outPoint)
			EndIf
		EndIf
			
		Return retval
	EndMethod
EndType

' detects if an ellipsoid intersects with a triangle of a polygon
Type DETECT_ELLIPSOID_POLY_IRRLICHT Extends ICOLLISION_DETECLICHT
	
	Function Create:DETECT_ELLIPSOID_POLY_IRRLICHT()
		Local retval:DETECT_ELLIPSOID_POLY_IRRLICHT=New DETECT_ELLIPSOID_POLY_IRRLICHT
		retval._method=CD_ELLIPSOID_POLY_IRRLICHT
		Return retval
	EndFunction
	
	' this is a ellipsoid-poly collision
	Method GetCollision:COLLISION(entity1:ib3d_Entity,entity2:ib3d_Entity)
	
		If Not MESH(entity2) Then RuntimeError "Cannot detect ELLIPSOID to POLY collisions unless checking against a MESH"
		If Not MESH(entity2)._mesh Then RuntimeError "Invalid MESH"

		Local retval:COLLISION=Null
		Local vel:Vector3df=entity1._node.getPosition().Minus(entity1.getLastPosition())

		' make sure we are moving
		If vel.Plus(_gravity).NotEq(Vector3df.createFromVals(0,0,0))
			' if this is an animated, build dynamically based on the current frame	
			Local triselector:ITriangleSelector=Null
				
			' for ellipsoid-poly, create the triangle selector each time
			If IAnimatedMesh(MESH(entity2)._mesh)
				triselector=_g_ib3d_engine.smgr.createTriangleSelector(IAnimatedMesh(MESH(entity2)._mesh).getMesh(IAnimatedMeshSceneNode(entity2._node).getFrameNr()),entity2._node)
			Else
				triselector=_g_ib3d_engine.smgr.createTriangleSelector(IMesh(MESH(entity2)._mesh),entity2._node)
			EndIf 
			
			Local pos:Vector3df		
			Local tri:Triangle3df=Triangle3df.Create()
			Local outPoint:Vector3df=Vector3df.Create()
			Local falling:Int=False
			Local outNode:ISceneNode
					
			pos=_g_ib3d_engine.smgr.getSceneCollisionManager().getCollisionResultPosition(triselector,entity1.getLastPosition().Minus(entity1._collision_offset),entity1._radius,vel,tri,outPoint,Varptr(falling),outNode,_slide,_gravity)
			pos.PlusEq(entity1._collision_offset)
							
			If falling Or (tri.getPointA().getX()<>0 Or tri.getPointA().getY()<>0 Or tri.getPointA().getZ()<>0)
				Return COLLISION.Create(entity1,entity2,_method,tri,pos,falling,outPoint)
			EndIf
		EndIf	
					
		Return retval
	EndMethod
EndType

' detects if an ellipsoid intersects with a triangle of a polygon made out of the bounding box of a node
Type DETECT_ELLIPSOID_BOX_IRRLICHT Extends ICOLLISION_DETECLICHT
	
	Function Create:DETECT_ELLIPSOID_BOX_IRRLICHT()
		Local retval:DETECT_ELLIPSOID_BOX_IRRLICHT=New DETECT_ELLIPSOID_BOX_IRRLICHT
		retval._method=CD_ELLIPSOID_BOX_IRRLICHT
		Return retval
	EndFunction
	
	' this is a sphere-poly collision
	Method GetCollision:COLLISION(entity1:ib3d_Entity,entity2:ib3d_Entity)
	
		If Not MESH(entity2) Then RuntimeError "Cannot detect SPHERE to POLY collisions unless checking against a MESH"
		If Not MESH(entity2)._mesh Then RuntimeError "Invalid MESH"

		Local retval:COLLISION=Null
		Local vel:Vector3df=entity1._node.getPosition().Minus(entity1.getLastPosition())

		' make sure we are moving
		If vel.Plus(_gravity).NotEq(Vector3df.createFromVals(0,0,0))
			' builds based on the current bounding box
			Local triselector:ITriangleSelector=_g_ib3d_engine.smgr.createTriangleSelectorFromBoundingBox(entity2._node)
						
			Local pos:Vector3df		
			Local tri:Triangle3df=Triangle3df.Create()
			Local outPoint:Vector3df=Vector3df.Create()
			Local falling:Int=False
			Local outNode:ISceneNode
					
			pos=_g_ib3d_engine.smgr.getSceneCollisionManager().getCollisionResultPosition(triselector,entity1.getLastPosition().Minus(entity1._collision_offset),entity1._radius,vel,tri,outPoint,Varptr(falling),outNode,_slide,_gravity)
			pos.PlusEq(entity1._collision_offset)
				
			'If Not Vector3df_Precision_EqEq(pos,entity1._node.getPosition())
			If falling Or (tri.getPointA().getX()<>0 Or tri.getPointA().getY()<>0 Or tri.getPointA().getZ()<>0)
				Return COLLISION.Create(entity1,entity2,_method,tri,pos,falling,outPoint)
			EndIf
		EndIf	
					
		Return retval
	EndMethod
EndType

' TODO: BOX BOX not functional yet
' uses the bounding boxes to determine if a collision has occurred
Type DETECT_BOX_BOX_IRRLICHT Extends ICOLLISION_DETECLICHT
	
	Function Create:DETECT_BOX_BOX_IRRLICHT()
		Local retval:DETECT_BOX_BOX_IRRLICHT=New DETECT_BOX_BOX_IRRLICHT
		retval._method=CD_BOX_BOX_IRRLICHT
		Return retval
	EndFunction
	
	' this is a sphere-shpere collision
	Method GetCollision:COLLISION(entity1:ib3d_Entity,entity2:ib3d_Entity)
		Local retval:COLLISION=Null
		
		Local box:AABBox3df=entity1._node.getTransformedBoundingBox()
		' if the box is empty we need to create one based on the radius
		If box.IsEmpty() Then
			Local vmin:Vector3df=Vector3df.createFromVect(entity1._radius)
			vmin.invert()
			box.resetFromBox(AABBox3df.createFromVectWithMax(vmin,entity1._radius))
			entity1._node.getAbsoluteTransformation().TransformBox(box)
			DebugLog("a: "+box.getMaxEdge().getX()+","+box.getMaxEdge().getY()+","+box.getMaxEdge().getZ())
			
		EndIf
		Local box2:AABBox3df=entity2._node.getTransformedBoundingBox()
		DebugLog("2: "+box2.getMaxEdge().getX()+","+box2.getMaxEdge().getY()+","+box2.getMaxEdge().getZ())
		
		
		' if the distance between is less than the sum of the radius', we have a collision
		'If box.intersectsWithBox(entity2._node.getTransformedBoundingBox())
		If entity2._node.getTransformedBoundingBox().intersectsWithBox(box)
			Return COLLISION.Create(entity1,entity2,_method)
		EndIf
		
		Return retval
	EndMethod
EndType


